# -*- perl -*-

package Common;

sub initfh {
    my $self = shift;
    my $fd = $self->{fd};
    
    $self->set_nbio(1);
    setsockopt( $fd, ::SOL_SOCKET(), ::SO_SNDBUF(), 64 * 1024 );
    setsockopt( $fd, ::SOL_SOCKET(), ::SO_RCVBUF(), 64 * 1024 );
    setsockopt( $fd, ::SOL_SOCKET(), ::SO_SNDLOWAT(), 8 * 1024 );
    
    $::files[ fileno($fd) ] = $self;
}

sub closeit {
    my $self = shift;
    my( $fh,  $x );

    return 0 if $me->{dead};
    
    $fh = $self->{fd};
    undef $::files[ fileno($fh) ];
    $x = close $fh;
    $self->{fd} = undef;
    $self->{dead} = 1;
    $self->{timeout} = 0;
    $x;
}
   
sub shutdown {
    my $self = shift;
    my( $fh );

    return 0 if $me->{dead};
    
    $fh = $self->{fd};
    shutdown( $fh, 2 );
    $self->closeit();
}

sub set_nbio {
    my($self, $mode) = @_;
    my $fh;

    $fh = $self->{fd};

    if( 1 ){
	
	my $curr = fcntl($fh, ::F_GETFL(), 0);
	$curr = 0 if $curr =~ /but true/;
	
	if($mode){
	    $curr |= ::O_NDELAY();
	}else{
	    $curr &= ~::O_NDELAY();
	}
	fcntl($fh, ::F_SETFL(), $mode);
    }else{
	ioctl($fh, ::FIONBIO(), $mode?1:0);
    }
}

sub write {
    my $me = shift;
    my $msg = shift;
    
    $me->{wbuffer} .= $msg;
    $me->{wantw} = 1;

    #    if( length($me->{wbuffer}) > 64000 ){
#	$me->shutdown();
#	$me->cleanup();
#    }
}

$WRSZ = 1024 * 64;
sub writable {
    my $me = shift;
    
    $fh = $me->{fd};

    $me->{debug} .= " w" if $me->{debug};
    
    if( $me->{wbuffer} ){
        my( $i, $l );

        $l = length($me->{wbuffer});
	$l = $WRSZ if $l > $WRSZ;
	#print STDERR "common:writable: ($l)\n";
	alarm(2);
        $i = syswrite $fh, $me->{wbuffer}, $l;
#	printf ::LOGW "%s\n", GNT::guid_ntoa( substr($me->{wbuffer}, 0, $i)) if $me->{mode} eq "server";
        $me->{wbuffer} = substr($me->{wbuffer}, $i);

	# print STDERR "common:wrote ($i of $l) to $me->{addr}\n";
    }

    if( $me->{wbuffer} ){
        $me->{wantw} = 1;
    }else{
        $me->{wantw} = 0;
        if( $me->{closeme} ){
            $me->shutdown();
	    $me->cleanup();
        }
    }
}

$myip = '';
sub myip {
    $me = shift;
    my( $ip, $s );
    
    if( !$myip ){
	$s = getsockname( $me->{fd} );
	if( $s ){
	    ($port, $ip) = ::unpack_sockaddr_in($s);
	}else{
	    print STDERR " $me->{addr} " if $::opt_D;
	}
	$myip = $ip unless( ::inet_ntoa($ip) eq "127.0.0.1" );
    }else{
	$ip = $myip;
    }
    $ip;
}


1;

