# -*- perl -*-

package Listen;

@ISA = qw(Common);

$verbose = 1;
$PORT;

sub new {
    my $class = shift;
    my $port  = shift;
    my $me = {};
    my( $fh, $i );

    bless $me, $class;
    $fh = ::anon_fh();
    $me->{fd} = $fh;
    $me->{name} = "listener";
    $i = socket($fh, ::PF_INET(), ::SOCK_STREAM(), getprotobyname('tcp'));
    $i = setsockopt($fh, ::SOL_SOCKET(), ::SO_REUSEADDR(), 1);
    $i = bind($fh, ::sockaddr_in($port, ::INADDR_ANY()));

    if( ! $i ){
	::loggit( "bind tcp/$port: $!" );
	exit;
    }
    
    listen($fh, 10);
    
    $me->initfh();
    $me->{wantr} = 1;
    $me->{wantw} = 0;
    $me->{timeout} = 0;
    $me->{port}  = $PORT = $port;

    ::loggit( "restarted. listening on tcp/$port" );

    $me;
}

sub readable {
    my $me = shift;
    my( $fh, $nfh, $c, $i );

    $fh  = $me->{fd};
    $nfh = ::anon_fh();
    accept( $nfh, $fh );
    
    $c = Conn->new( ::inet_ntoa( (::sockaddr_in(getpeername($nfh)))[1] ), $me->{port} );
    $c->{fd} = $nfh;
    $c->initfh();
    $c->server();
}

################################################################

package Conn;

@ISA = qw(Common);

@all = ();
@pending = ();

# everything is a Conn, until the handshake determines whether it
# is a gnutella or http connection

sub new {
    my $class = shift;
    my $addr = shift;
    my $port = shift;
    my $me = {};
    bless $me, $class;

    $me->{addr} = $addr;
    $me->{port} = $port;
    $me->{time} = $^T;
    $me->{wbuffer} = '';
    $me->{rbuffer} = '';
    $me->{wantw} = 0;
    $me->{wantr} = 0;
    $me->{timeout} = 0;

    $me;
}

sub server {
    my $me = shift;

    $me->{mode} = "server";
    $me->{wantr} = 1;
    $me->{timeout} = $^T + 120;
    push @all, $me;
    
    # RSN - check ACL?
}

sub scheduleclient {
    my $who  = shift;
    my $when = shift;
    my( $d, $ph );

    $d = $when - $^T;
    $ph = rand($d/5) - $d/10;
    $when += $ph;
    push @pending, [ $when, $who ];
}

sub startpending {

    while( @pending && $pending[0][0] <= $^T ){
	Conn->startnewclient( $pending[0][1] );
	shift @pending;
    }
    if( @pending ){
	return $pending[0][0];
    }else{
	return 0;
    }
}

sub startnewclient {
    my $class = shift;
    my $who   = shift;
    my $me;
    my( $addr, $port, $ip );

    print STDERR "start: $who\n" if $::opt_v;
    $me = $class->new();
    ($addr, $port) = split /:/, $who;
    $me->{addr} = $addr;
    $me->{port} = $port;

    push @all, $me;
    $me->startclient();
}

sub startclient {
    my $me = shift;
    my( $i, $fh, $ip );

    $fh = ::anon_fh();
    $ip = ::resolve($me->{addr});
    if( !$ip ){
	print STDERR "cannot resolve $addr\n";
	return;
    }

    $i = socket($fh, ::PF_INET(), ::SOCK_STREAM(), getprotobyname('tcp'));
    # print STDERR "sc: $i $!\n";
    $me->{fd} = $fh;
    $me->initfh();
    alarm(2);
    $i = connect( $fh, ::sockaddr_in($me->{port}, $ip) );
    alarm(0);
    #print STDERR "sc: $i $!\n";
    $me->{wantr}   = 1;
    $me->{wantw}   = 1;
    $me->{timeout} = $^T + 120;
    $me->{mode}    = "client";
    $me->write( "GNUTELLA CONNECT/0.4\n\n" );
    # $me->{debug} = "startclient";
}

sub timeout {
    my $me = shift;

    $me->{debug} .= " TO" if $me->{debug};
    $me->shutdown();
    $me->cleanup();
}

sub cleanup {
    my $me = shift;

    @all = grep { $_ != $me } @all;
    if( $me->{mode} eq "client" ){
        scheduleclient( "$me->{addr}:$me->{port}", $^T + $::RESCHED );
    }
}

sub readable {
    my $me = shift;
    my( $i, $l, $fh );

    $fh = $me->{fd};

    # QQQ or just read 1 -- will simplify the code ...
    $i = sysread $fh, $l, 8192;
    $me->{debug} .= " r$i" if $me->{debug};
    
    # print STDERR "conn:r: ($i) <$l>\n";
    if( $i ){
	$me->{rbuffer} .= $l;
	if( $me->{rbuffer} =~ /\n\n|\r\n\r\n/ ){
	    @all = grep { $_ != $me } @all;
	    if( $me->{rbuffer} =~ /^gnutella/i ){
	        GNT->rebless( $me );
	    }
	    elsif( $me->{rbuffer} =~ /^get|head/i ){
	        HTTP->rebless( $me );
	    }
	    else{
		# unknown ?
		$me->write( "WTF?\n" );
		$me->{closeme} = 1;
	    }
	}
    }else{
	# eof
	$me->shutdown();
	@all = grep { $_ != $me } @all;

	if( $me->{mode} eq "client" ){
	  Conn::scheduleclient( "$me->{addr}:$me->{port}", $^T + $::RESCHED );
	}

    }
}

1;
