# -*- perl -*-

package GNT;

@ISA = qw(Common);

$RT_TO = 300;
$MAX_TABLE = 100000;

$MAX_BUFFER_HARD = 64000;
$MAX_BUFFER_SOFT = int($MAX_BUFFER_HARD * 7 / 8);

# all connections
@all = ();

# routing table
@rtbytime = ();
%rtbyguid = ();

# stats
%rcvstats = ();
%sndstats = ();

sub rebless {
    my $class = shift;
    my $me = shift;

    bless $me, GNT;
    if( $me->{mode} eq "client" ){
	$me->{rbuffer} =~ s/.*\n\n//;
    }

#    print STDERR "-> $me->{rbuffer}\n";

    if( $me->{mode} eq "server" ){
	if( @all + @Conn::all + @Conn::pending >= $::MAX_CONNS ){
	    $me->write( "maximum number of connections exceeded. I just can't handle it. Arrrrrg!\n\n" );
	    $me->{closeme} = 1;
	    ::loggit( "connection refused from $me->{addr} - maximum limit reached" );
	    return;
	}
	if( $::hostdeny{$me->{addr}} ){
	    $me->write( "access denied. sorry\n\n" );
	    $me->{closeme} = 1;
	    ::loggit( "connection refused from $me->{addr} - is on deny list" );
	    return;
	}
	$me->write( "GNUTELLA OK\n\n" );
	$me->{rbuffer} = '';
    }

    $me->{wantr}   = 1;
    $me->{timeout} = 0;
    $me->{dropped} = 0;
    ::loggit( "gnutella connection established: $me->{addr} ($me->{mode})" );
    push @all, $me;

    $me->procrbuf() if( $me->{rbuffer} );
}

# remove from all, resched
sub cleanup {
    my $me = shift;

    print STDERR "*" if $::opt_D;
    if( $me->{mode} eq "client" ){
        Conn::scheduleclient( "$me->{addr}:$me->{port}", $^T + $::RESCHED );
    }

    @all = grep { $_ != $me } @all;

    ::loggit( "gnutella connection terminated: $me->{addr} - in=$me->{pktin} out=$me->{pktout} drp=$me->{dropped}" );
}

$RDSZ = 64 * 1024;
sub readable {
    my $me = shift;
    my( $fh );

    $fh = $me->{fd};

    alarm(2);
    $::reads_total ++;
    $i = sysread $fh, $l, $RDSZ;
    if( $i == $RDSZ ){
	$::reads_full ++;
	print STDERR "#" if $::opt_D;
    }
    if( $i > 0 ){
        $me->{rbuffer} .= $l;
    }
    if( $i < 0 ){
	if( $! == 4 ){
	    # eintr
	}else{
	    ::loggit( "gnutella connection read error: $me->{addr} - $!" );
	    $me->cleanup();
	    $me->shutdown();
	}
    }
    
    if( ! $i ){
        # eof
	$me->cleanup();
	$me->shutdown();
    }

    $me->procrbuf() if( $me->{rbuffer} );
}

sub procrbuf {
    my $me = shift;
    my( $len, $dlen );

    $len = length( $me->{rbuffer} );

    while( $len >= 23 && $len >= packet_length($me->{rbuffer}) + 23 ){
	my( $dlen, $pkt );
	return if $me->{dead};
	$dlen = packet_length($me->{rbuffer});
	if( $dlen > 100000 || $dlen < 0 ){
	    $me->shutdown();
	    $me->cleanup();
	    print STDERR "[$dlen]" if $::opt_D;
	    return;
	}
	$pkt = substr( $me->{rbuffer}, 0, $dlen + 23 );
	$me->{rbuffer} = substr( $me->{rbuffer}, $dlen + 23 );
	$len = length( $me->{rbuffer} );

	$me->handle_packet( $pkt );
    }
}

# should not happen
sub timeout {
    my $me = shift;
    
    print STDERR "jeff says this should not happen\n";
    $me->cleanup();
    $me->shutdown();
}

sub packet_length {
    my $pkt = shift;

    little_endian( substr($pkt, 19, 2) );
}

sub dumppkt {
    my $pkt = shift;
    
    printf STDERR "\t%d<%s.%s>\n", length($pkt), substr(guid_ntoa($pkt), 0, 32),substr(guid_ntoa($pkt), 32);
}

sub handle_packet {
    my $me  = shift;
    my $pkt = shift;
    my( $func, $x );
    
    # update ttl, hops
    substr( $pkt, 17, 1 ) = chr(ord(substr( $pkt, 17, 1 ))-1);
    substr( $pkt, 18, 1 ) = chr(ord(substr( $pkt, 18, 1 ))+1);

    my( $mid, $funcn, $ttl, $hops, $len )
        = unpack( "a16 C C C a2", $pkt );

    $len  = little_endian( $len );
    $mid  = guid_ntoa($mid);
    $func = pkt_type($funcn);

    # dumppkt($pkt);

    # validate packet
    if( $func eq "ping" && $len != 0){
	::loggit( "ping != $len\n" );
	$me->{errors} ++;
	$rcvstats{errors} ++
    }
    elsif( $func eq "pong" && $len != 14 ){
	::loggit( "pong != $len\n" );
	$me->{errors} ++;
	$rcvstats{errors} ++;
	return if $len < 14;
    }
    elsif( !$func ){
	$me->{errors} ++;
	$rcvstats{errors} ++;
	$func = "????";
    }
    else{
	$me->{errors} = 0;
    }

    # keep some stats
    $statshops[ $hops ] ++;
    $statssize[ $len / 8 ] ++;
    $me->{pktin} ++;
    $rcvstats{totalpkts} ++;
    $rcvstats{$func } ++;

    return if $ttl < 1 || $ttl > 50 || $hops > 50;

    if( $me->{errors} > 5 ){
	::loggit( "dropping connection with $me->{addr} - too many errors" );
	$me->cleanup();
	$me->shutdown();
    }

    #print STDERR "$func $ttl $hops $len $mid\n";
    print STDERR substr( $func, 1, 1) if $::opt_D;

    if( $funcn & 1 ){
	# route
	$x = findroute( $mid );
	if( !$x ){
	    $rcvstats{noroute} ++;
	    return;
	}
	return if $x->{dead};
	$x->pktoutput( $pkt );
	$sndstats{$func || "#$funcn"} ++;
	$sndstats{totalpkts} ++;
    }else{
	# broadcast
	# seen it already?
	if( $rtbyguid{$mid} ){
	    $rcvstats{dupe} ++;
	    return;
	}
	foreach $x (@all){
	    next if $x == $me;
	    next if $x->{dead};
	    # site cannot keep up--start by dropping bcast pkts
	    if( length($x->{wbuffer}) >= $MAX_BUFFER_SOFT ){
		$x->{dropped} ++;
		next;
	    }
	    $x->pktoutput( $pkt );
	    $sndstats{$func || "#$funcn"} ++;
	    $sndstats{totalpkts} ++;
	}
	addroute( $me, $mid );
	$me->pong($mid) if( $func eq "ping" );
	$me->search($mid, $pkt) if( $func eq "srch" );
    }
}

sub pktoutput {
    my $me = shift;
    my $pkt = shift;

    return if length($pkt) > 64000;
    if( length($me->{wbuffer}) > $MAX_BUFFER_HARD ){
	$me->{dropped} ++;
	return;
    }
    $me->{pktout} ++;
    $me->write( $pkt );

}


################################################################

# reply to ping
sub pong {
    my $me = shift;
    my $guid = shift;
    my( $port, $pkt, $r, $s );

    return if $::NO_PONG;
    $sndstats{ponged}++;

    $r = pack("a2 a4 a4 a4",
	      to_little_endian2( $Listen::PORT ),
	      $me->myip(),
	      # files, no bytes...
	      to_little_endian4( @::filesbyindex ),
	      "\0\0\0\0" );
   
    $pkt = guid_aton($guid) . pack("C C C a4", 1, 7, 0, to_little_endian4(length($r))) . $r;

    $me->pktoutput( $pkt );
    print STDERR "O" if $::opt_D;
}

################################################################

# search our files, and reply
sub search {
    my $me = shift;
    my $guid = shift;
    my $pkt = shift;
    my( $msp, $qry, $w, @w, $r, @r, %r );

    $pkt = substr( $pkt, 23 );
    $msp = unpack( "A2", $pkt );
    $msp = little_endian( $msp );
    ($qry) = $pkt =~ /^..(.*)\0/;

    return unless $::MY_SPEED >= $msp;

    @w = ::q2w($qry);

    foreach $w (@w){
	map {$r{$_}++} @{$::indexbyword{$w}};
	# RSN ? keep stats {$w}++
    }

    foreach $r (keys %r){
	push @r, $r if( $r{$r} == @w );
    }

    if( @r ){
	my $newpkt = pack( "C a2 a4 a4",
			   scalar(@r),
			   to_little_endian2($Listen::PORT),
			   $me->myip(),
			   to_little_endian4($::MY_SPEED) );
	foreach $r (sort {$a <=> $b} @r){
	    my( $size, $name, $pname );

	    $name = $::filebyindex[$r][2];
	    $size = $::filebyindex[$r][1];

	    $newpkt .= pack( "a4 a4",
			     to_little_endian4( $r ),
			     to_little_endian4( $size ) );
	    $newpkt .= "$name\0\0";
	}

	$newpkt .= "\0" x 16;	# Bite me!
	$pkt = guid_aton($guid) . pack("C C C a4", 0x81, 7, 0, to_little_endian4(length($newpkt))) . $newpkt;

	$sndstats{found}++;
	$me->pktoutput( $pkt );

	print STDERR "R" if $::opt_D;

	# XXX stats
    }

}

################################################################

sub send_query {
    my $who   = shift;
    my $query = shift;
    my( $x, $p, $q, $g );

    $g = make_guid();
    $q = "\0\0$query\0";
    $p = guid_aton($g) . pack "C C C a4", 128, 7, 0,
      to_little_endian4(length($q));
    $p .= $q;

#    print STDERR "sq: $query <$g>\n";
#    printf STDERR "\t%d<%s.%s>\n", length($p), substr(guid_ntoa($p), 0, 32),substr(guid_ntoa($p),32),;
    
    # add route
    addroute( $who, $g );
    
    # bcast
    foreach $x (@all){
	next if $x->{dead};
	$x->pktoutput( $p );
	$sndstats{srch} ++;
	print STDERR "Q" if $::opt_D;
    }

    # QQQ - should I reply to myself with my own files?
    search($who, $g, $p);
}

sub send_ping {
    my $who   = shift;
    my( $x, $p, $g );

    $g = make_guid();
    $p = guid_aton($g) . pack "C C C a4", 0, 7, 0, 0;

    # add route
    addroute( $who, $g );
    
    # bcast
    foreach $x (@all){
	next if $x->{dead};
	$x->pktoutput( $p );
	$sndstats{ping} ++;
	print STDERR "P" if $::opt_D;
    }
}


sub make_guid {
    my( $g );

    $g .= sprintf("%x",rand(16)) foreach (0..31);
    $g;
}

################################################################

sub findroute {
    my $guid = shift;
    my( $x );

    $x = $rtbyguid{$guid};
    return undef unless $x;
    return $x->{conn};
}

sub addroute {
    my $who = shift;
    my $guid = shift;
    my( $e );

    $e = {
	conn => $who,
	time => $^T,
	guid => $guid,
    };

    $rtbyguid{$guid} = $e;
    push @rtbytime, $e;
}

sub rtcleanup {
    my $x;

    while( (@rtbytime && $rtbytime[0]{time} + $RT_TO < $^T)
	   || (@rtbytime > $MAX_TABLE) ){
	$x = $rtbytime[0]{guid};
	shift @rtbytime;
	delete $rtbyguid{$x};
	# print STDERR "X"
    }


}

################################################################

sub little_endian {
    my $n = shift;
    my $a = 0;

    map {$a<<=8; $a+=ord($_);  } reverse split('', $n);
    $a;
}

sub to_little_endian2 {
    my $n = shift;

    pack( "CC", $n&0xFF, ($n>>8)&0xFF );
}

sub to_little_endian4 {
    my $n = shift;

    pack( "CCCC", $n&0xFF, ($n>>8)&0xFF, ($n>>16) & 0xFF, ($n>>24)&0xFF );
}


sub guid_ntoa {
    my $m = shift;
    
    join( '', (map {sprintf "%0.2x", ord($_)} split( '', $m) ));
}

sub guid_aton {
    my $g = shift;
    my($a, @a);

    while($g){
	($a,$g) = $g =~ /^(..)(.*)/;
	push @a, $a;
    }
    join('', (map {chr(hex($_))} @a));
}

sub pkt_type {
    my $n = shift;

    { '0' => "ping",
      '1' => "pong",
      '64' => "push",
      '128' => "srch",
      '129' => "rslt",
      }->{ $n };
}


1;

