# -*- perl -*-

package HTTP;

@ISA = qw(Common);

$verbose = 0;
$dnlds   = 0;
$lastping = 0;

$PINGDT   = 120;

sub header {
    my $me  = shift;
    my $msg = shift;
    my $type = shift || "text/html";
    my @headers = @_;

    $me->write( $me->headertxt($msg, $type, @headers) );
    $me->write( "\r\n" );
}

sub headertxt {
    my $me  = shift;
    my $msg = shift;
    my $type = shift || "text/html";
    my @headers = @_;
    my( $b );
    
    $b = "HTTP/1.0 $msg\r\n" .
	 "Date: " . scalar(localtime) . "\r\n" .
	 "Server: peagnutbutter (Jeff's Gnutella Program) - Vers. A\r\n" .
	 "Pragma: no-cache\r\n" .
	 "Cache-control: private\r\n" .
	 "Content-type: $type\r\n";

    foreach (@headers){
	$b .= "$_\r\n";
    }
    $b .= "\r\n";
}


sub badrequest {
    my $me  = shift;
    my $msg = shift;
    my $details = shift;
    
    $me->header("$msg");
    $me->write( "<HEAD><TITLE>$msg</TITLE></HEAD>\n<BODY>\n<H1>$msg</H1>\n" .
		( $details ? "$details\n" : "" ) . "</BODY>\n" );
    $me->{closeme} = 1;
}

sub startpage {
    my $me = shift;
    my $title = shift;
    
    $me->header( "200 Document follows" );
    $me->write( "<HEAD><TITLE>$title</TITLE></HEAD>\n<BODY BGCOLOR=#FFFFFF>\n" );
}

sub endpage {
    $me = shift;

    # $me->write( "<HR>\n<A HREF=\"http://www.op.net/~jaw/pgb/\"><I>peagnutbutter</I></A>\n" );
    $me->write( "<HR>\n<A HREF=\"http://www.tcp4me.com/code/pgb/\"><I>peagnutbutter</I></A>\n" );
    $me->write( "</BODY>\n" );
}

sub rebless {
    my $class = shift;
    my $me = shift;
    my( $meth, $file, $request, $x, $d, $foo );

    
    bless $me, $class;
    $request = $me->{rbuffer};
    ($meth, $file) = $request =~ /(get)\s+(.*)\s+HTTP/i;
    print STDERR "HTTP: $meth $file\n" if $verbose;
    if( !$meth || !$file ){
	$me->badrequest( "400 Bad Request" );
	$request =~ s/[\r\n\t]/ /g;
	::loggit( "$me->{addr} HTTP 400 " . substr($request, 0, 64) );
	return;
    }
    $me->{rbuffer} = '';
    $me->{wantr}   = 0;

    # RSN - list all of our files, ping/pong
    if( $file eq "/connections" ){
	::loggit( "$me->{addr} HTTP 200 $meth $file" );
	$me->startpage("Connections");
	$me->write( "<TABLE BORDER=1>\n" );
	$me->write( "<TR><TH>site</TH><TH>mode</TH><TH>connect time</TH><TH>pkt in</TH><TH>pkt out</TH><TH>curr. buflen</TH><TH>drops</TH></TR>\n" );
	foreach $x (@GNT::all){
	    $d = localtime( $x->{time} );
	    $sz = length( $x->{wbuffer} );
	    $me->write( "<TR><TD>$x->{addr}</TD><TD>$x->{mode}</TD><TD>$d</TD><TD>$x->{pktin}</TD><TD>$x->{pktout}</TD><TD>$sz</TD><TD>$x->{dropped}</TD></TR>\n" );
	}

	if( $::opt_D ){
	    foreach $x (@Conn::all){
		$d = localtime( $x->{time} );
		$sz = length( $x->{wbuffer} );
		$me->write( "<TR><TD>$x->{addr}</TD><TD>($x->{mode})</TD><TD>$d</TD><TD COLSPAN=4>" );
		$d = $x->{timeout} ? localtime( $x->{timeout} ) : "never";
		$me->write( "r=$x->{wantr}/w=$x->{wantw}/s=$sz/$d $x->{debug}</TD></TR>\n" );
	    }
	}
	$me->write( "</TABLE>\n" );
	$me->endpage();
    }

    elsif( $file eq "/pending" ){
	::loggit( "$me->{addr} HTTP 200 $meth $file" );
	$me->startpage("Pending");
	$me->write( "<TABLE BORDER=1>\n" );

	foreach $x (@Conn::pending){
	    $d = localtime( $x->[0] );
	    $me->write( "<TR><TD>$x->[1]</TD><TD>$d</TD></TR>\n" );
	}
	$me->write( "</TABLE>\n" );
	$me->endpage();
    }
	    
    elsif( $file eq "/stats" ){
	::loggit( "$me->{addr} HTTP 200 $meth $file" );
	$me->startpage("Stats");
	$me->write( "<TABLE BORDER=1>\n" );
	$me->write( "<TR><TH>type</TH><TH>rvcd</TH><TH>sent</TH></TR>\n" );
#	foreach $x (sort keys %GNT::rcvstats){
	foreach $x (qw(ping pong srch rslt push ???? dupe noroute ponged found)){
	    $me->write( "<TR><TD>$x</TD><TD>$GNT::rcvstats{$x}</TD><TD>$GNT::sndstats{$x}</TD></TR>\n" );
	}

	$me->write( "<TR><TD COLSPAN=3><B>number of packets by distance</TD></TR>\n" );
	foreach $x (0 .. $#GNT::statshops){
	    next unless $GNT::statshops[$x];
	    $me->write( "<TR><TD>$x hops</TD><TD>$GNT::statshops[$x]</TD><TD><BR></TD></TR>\n" );
	}
	$me->write( "<TR><TD COLSPAN=3><B>number of packets by size (in 8 bytes)</TD></TR>\n" );
	foreach $x (0 .. $#GNT::statssize){
	    next unless $GNT::statssize[$x];
	    $me->write( "<TR><TD>$x quads</TD><TD>$GNT::statssize[$x]</TD><TD><BR></TD></TR>\n" );
	}

	$me->write( "</TABLE>\n" );
	$me->endpage();
	
    }

    elsif( $file eq "/menu" ){
	::loggit( "$me->{addr} HTTP 200 $meth $file" );
	$me->startpage("Menu");
	$me->write( <<EOI
		    <LI><A HREF="/form">Search for files</A>
		    <LI><A HREF="/ping">Search for other hosts</A>
		    <LI><A HREF="/connections">List of current connections</A>
		    <LI><A HREF="/pending">List of pending connections</A>
		    <LI><A HREF="/stats">Statistics</A>
EOI
		    );
	
    	$me->endpage();
    }

    elsif( $file eq "/form" ){
	::loggit( "$me->{addr} HTTP 200 $meth $file" );
	$me->startpage("Query Form");
	$me->write( <<EOI
		    <FORM ACTION="/query" method="GET">
		    Max Results:
		    <INPUT TYPE="text" VALUE="500" NAME="results" SIZE="4">
		    Timeout:
		    <INPUT TYPE="text" VALUE="300" NAME="timeout" SIZE="3"> (seconds)<BR>
		    Search for:
		    <INPUT TYPE="text" VALUE="" NAME="query" SIZE="30"><BR>
		    <INPUT TYPE="submit" VALUE="submit">
		    </FORM>
EOI
		    );
	
	$me->endpage();
    }

    elsif( $file =~ /^\/ping(.*)/ ){
	my( $foo, $to, $max );

	$foo = $1;
	$foo =~ s/\+/ /g;
	$foo =~ s/%20/ /g;
	
	if( $foo =~ /^\?/ ){
	    ($max)   = $foo =~ /max=([^\?\&]*)/;
	    ($to)    = $foo =~ /timeout=([^\?\&]*)/;
	}else{
	    $max     = 0;
	    $to      = 240;
	}

	# rate limit pings, so we don't flood the network
	if( $^T < $lastping + $PINGDT ){
	    ::loggit( "$me->{addr} HTTP 403 $meth $file (too soon to ping again)" );
	    $me->badrequest( "403 Not Permitted", "I'm sorry, this is so tiring. Maybe later." );
	    return;
	}
	    
	::loggit( "$me->{addr} HTTP 200 $meth $file" );
	$me->startpage("Ping");
	$me->write( "<H2>Results of Ping</H2>\n" );
	$me->{queryinprogress} = 1;
	$me->{timeout} = $to + $^T;
	$me->{maxresults} = $max;
        GNT::send_ping( $me );
	return;
    }
    
    elsif( $file =~ /^\/query(.*)/ ){
	my( $foo, $to, $max, $query, @q );

	$foo = $1;
	$foo =~ s/\+/ /g;
	$foo =~ s/%20/ /g;
	
	if( $foo =~ /^\/(.*)/ ){
	    $query = $1;
	    $to = 300;
	    $max = 1000;
	}elsif( $foo =~ /^\?/ ){
	    ($query) = $foo =~ /query=([^\?\&]*)/;
	    ($max)   = $foo =~ /results=([^\?\&]*)/;
	    ($to)    = $foo =~ /timeout=([^\?\&]*)/;
	}

	# print STDERR "Query: $query, $max, $to\n";
	
	::loggit( "$me->{addr} HTTP 200 $meth $file" );
	$me->startpage("Query");
	$me->{queryinprogress} = 1;
	$me->{query} = $query;
	$me->{timeout} = $to + $^T;
	$me->{maxresults} = $max;

	$query =~ s/%3B/\;/g;
	@q = split /\s*;\s*/, $query;
	foreach $query (@q){
	    $me->write( "<H2>Results of Query: <I>$query</I></H2>\n" );
	    GNT::send_query( $me, $query );
	}
	return;
    }

    elsif( $file =~ /^\/get\/([^\/]*)/ ){
	# send file
	# RSN - stats
	my $i = $1;
	my( $buf, $sz, $start, $end );

	if( $i >= @::filebyindex ){
	    ::loggit( "$me->{addr} HTTP 404 $meth $file (index out of range)" );
	    $me->badrequest( "404 Not Found", "I'm sorry Dave, I can't do that" );
	    return;
	}

	if( $::MAX_DNLDS && $dnlds >= $::MAX_DNLDS ){
	    ::loggit( "$me->{addr} HTTP 503 $meth $file (maximum limit reached)" );
	    $me->badrequest( "503 Server Error",
			     "maximum number of connections exceeded. I just can't handle it." );
	    return;
	}
	
	if( ! open(F, $::filebyindex[$i][0]) ){
	    my $e = $!;
	    ::loggit( "$me->{addr} HTTP 500 $meth $file (open failed: $e)" );
	    $me->badrequest( "500 Server Error", "I said 'open', the system replied '$e'" );
	    return;
	}

	$size  = $::filebyindex[$i][1];
	$start = 0;
	$end   = $size;
	if( $request =~ /Range:\s*(\d*)\s*-\s*(\d*)/ ){
	    $start = $1;
	    $end   = $2;
	    $start = 0 unless $start < $size;
	    $end   = $size unless $end > $start;
	    $size  = $end - $start;
	}
	
	::loggit( "$me->{addr} HTTP 200 $meth get/$i/$::filebyindex[$i][0] - $size/$::filebyindex[$i][1]" );
	
	if( $pid=fork ){
	    # parent
	    # print STDERR "child: $pid (me: $$)\n";
	    $dnlds ++;
	    $me->closeit();
	    close F;
	    $SIG{CHLD} = sub {
		while( ($pid=waitpid(-1, 1)) > 0 ){
		    $dnlds --;
		    # ::loggit( "$pid done" );
		}
	    };
	    return;
	}else{
	    # child
	    my $fh = $me->{fd};
	    my $tot = 0;
	    my $bs = 1024;
	    my( $st, $ct, $lt );
	    
	    $me->set_nbio(0);
	    
	    $0 = "pgb/http $me->{addr} $::filebyindex[$i][0]";
	    print $fh $me->headertxt("200 Data Follows",
				     "application/binary",
				     "Content-length: $size" );

	    seek(F, $start, 0) if $start;
	    $st = $lt = time();
	    while( 1 ){
		$bs = $size - $tot if $bs > $size - $tot;
		$n = read(F, $buf, $bs);
		last if $n <= 0;
                setsockopt( $fh, ::SOL_SOCKET(), ::SO_SNDBUF, $n );
		$m = print $fh $buf;
		last unless $m;
		$tot += $n;
		last if $tot == $size;
		$ct = time();

                if( $ct != $lt ){
                    $bps = sprintf "%.2f", 8 * $bs / ($ct - $lt) / 1024;
                    $pct = sprintf "%.1f", $tot * 100 / $size;
                    $0 = "pgb/http $me->{addr} $pct% ${bps}kbps";
                }
                $bs *= 2 if( $ct - $lt < 5 );
                $lt = $ct;
	    }
            ::loggit("$me->{addr} get/$i/$::filebyindex[$i][0] - $tot bytes at $bps kbps" );
	    close F;
	    $me->shutdown();
	    exit;
	}

    }
    
    else{
	::loggit( "$me->{addr} HTTP 404 $meth $file" );
	$me->badrequest( "404 Not Found", "I see no $file here." );
    }
    
    $me->{closeme} = 1;
}

sub readable {
    my $me = shift;

    # eof?

    $me->{closeme} = 1;
    $me->shutdown();
    $me->cleanup();
}

sub cleanup {
    my $me = shift;

}

sub timeout {
    my $me = shift;

    # print STDERR "http to\n";
    $me->{timeout} = 0;
    if( $me->{queryinprogress} ){
	if( ! $me->{nfound} ){
	    $me->write( "Nothing Found<BR>\n" );
	}else{
	    $me->write( "<HR>\n<B>Timeout!</B>\n" );
	}
	$me->endpage();
	$me->{closeme} = 1;
    }else{
	$me->shutdown();
	$me->cleanup();
    }
}

sub pktoutput {
    my $me = shift;
    my $pkt = shift;

    my( $mid, $funcn, $ttl, $hops, $len )
        = unpack( "a16 C C C a2", $pkt );
    $mid  = GNT::guid_ntoa($mid);
    $func = GNT::pkt_type($funcn);
    $len  = GNT::little_endian( $len );

    if( $func eq "pong" ){

	$pkt = substr($pkt, 23);
	my( $prt, $ip, $fls, $sz ) = unpack( "a2 a4 a4 a4", $pkt );
	$prt = GNT::little_endian( $prt );
	$fls = GNT::little_endian( $fls );
	$sz  = GNT::little_endian( $sz  );
	$ip  = ::inet_ntoa( $ip );

	if( $sz >= 1000000 ){
	    $sz = sprintf "%.1fG", $sz/1000000;
	}else{
	    $sz = sprintf "%.1fM", $sz/1000;
	}
	$me->write( "<DT>$ip:$prt\n" );
	$me->write( "<DD>hops $hops\n" );
	$me->write( "<DD>files $fls\n" ) if $fls;
	$me->write( "<DD>size $sz\n" )   if $sz ne "0.0k";
	$me->write( "<HR>\n" );

	$me->{nfound} ++;
	    
	if( $me->{maxresults} && $me->{nfound} >= $me->{maxresults} ){
	    $me->write( "<HR>\nstopping after $me->{nfound} results<BR>\n" );
	    $me->endpage();
	    $me->{closeme} = 1;
	    return;
	}
    }
    
    elsif( $func eq "rslt" ){
	
	$pkt = substr($pkt, 23);
	my($num, $prt, $ip, $spd) = unpack( "C a2 a4 a4", $pkt );
	$prt = GNT::little_endian( $prt );
	$spd = GNT::little_endian( $spd );
	$ip  = ::inet_ntoa( $ip );
	
	return if( $ip =~ /^(0|10|192\.168|0|172\.(1[6789]|2.|3[01])\.)/ );
	return if( $::hostdeny{$ip} );
	
	$pkt = substr( $pkt, 11 );
	
	if( $spd >= 1000 ){
	    $spd = sprintf("%.1fM", $spd/1000);
	}else{
	    $spd = "${spd}k";
	}
	
	$me->write("<LI>$ip - $spd\n");
	for($i=0; $i< $num; $i++ ){
	    my( $name, $r, $extra );
	    my( $in, $sz ) = unpack( "a4 a4", $pkt );
	    $pkt = substr($pkt, 8);

	    ($r, $extra, $pkt) = split /\0/, $pkt, 3;
	    # ($r, $pkt) = split /\0\0/, $pkt, 2;
	    
	    $in = GNT::little_endian($in);
	    $sz = GNT::little_endian($sz);
	    
	    if( $sz >= 1000000 ){
		$sz = sprintf "%.1fM", $sz/1000000;
	    }else{
		$sz = sprintf "%.1fk", $sz/1000;
	    }
	    
	    # porno site lamers
	    if( $r eq " $me->{query}.html" ){
		::loggit( "porno lamer: adding $ip to deny list" );
		$::hostdeny{$ip} = 1;
		$me->write("<HR>\n");
		return;
	    }
	    
	    $name = $r;
	    $r =~ s/\s/%20/g;
	    # print STDERR "http://$ip:$prt/get/$in/$r\n";
	    $me->write( "<DD><A HREF=http://$ip:$prt/get/$in/$r TARGET=xxx>$name</A> - $sz\n" );

	    $me->{nfound} ++;
	    
	    if( $me->{maxresults} && $me->{nfound} >= $me->{maxresults} ){
		$me->write( "<HR>\nstopping after $me->{nfound} results<BR>\n" );
		$me->endpage();
		$me->{closeme} = 1;
		return;
	    }
	}
	$me->write("<HR>\n");
    }

    else{
	$me->write( "unknown packet type - $func<BR>\n" );
    }
}

1;

